/* KEYDRV.C
  The low-level keyboard driver for the SDL port.
  Functions a keyboard driver must provide are KeyDrv_Start() and
  KeyDrv_Stop(). Once KeyDrv_Start() is called the keyboard driver
  needs to somehow (platform-specific) keep immediate_keytable[]
  filled up with the current key states.
*/

#include "../keen.h"
#include "../keenext.h"
#include <SDL.h>

char KeyDrv_Start(void)
{
  VidDrv_printf("KeyDrv_Start()\n");
  return 0;
}

void KeyDrv_Stop(void)
{
}

// we process the cheat codes here so that it's not necessary to
// have a seperate keytable[] index for each cheat code letter
char key_g=0, key_o=0, key_d=0;

/* poll_events() is called from SpeedThrottle() (timedrv.c) */
void poll_events(void)
{
  char newState;
  SDL_Event event;
    if(SDL_PollEvent(&event))
    {
      switch( event.type ) 
        {
        case SDL_QUIT:
          crashflag = QUIT_NONFATAL;
          why_term_ptr = "SDL: Got quit event!";
          break;

        case SDL_KEYDOWN:
          newState = 1;
          goto prockey;
        case SDL_KEYUP:
          newState = 0;
prockey: ;
          switch(event.key.keysym.sym)
          {
           case SDLK_UP: immediate_keytable[KUP] = newState; break;
           case SDLK_DOWN: immediate_keytable[KDOWN] = newState; break;
           case SDLK_LEFT: immediate_keytable[KLEFT] = newState; break;
           case SDLK_RIGHT: immediate_keytable[KRIGHT] = newState; break;
           case SDLK_LCTRL: case SDLK_RCTRL:
                            immediate_keytable[KCTRL] = newState; break;
           case SDLK_LALT:  case SDLK_RALT:  
                            immediate_keytable[KALT] = newState; break;

           case SDLK_w:
                if (levelcontrol.curlevel==WORLD_MAP)
                {
                  immediate_keytable[KUP2] = newState;
                }
                else
                {
                  immediate_keytable[KCTRL2] = newState;
                  immediate_keytable[KUP2] = 0;
                }
                break;
           case SDLK_q: immediate_keytable[KALT2] = newState;
                        immediate_keytable[KQ] = newState;
                        break;
           case SDLK_s: immediate_keytable[KDOWN2] = newState; break;
           case SDLK_a: immediate_keytable[KLEFT2] = newState; break;
           case SDLK_d: immediate_keytable[KRIGHT2] = newState;
                        key_d = newState;
                        break;

           case SDLK_c: immediate_keytable[KC] = newState; break;

           case SDLK_SPACE: immediate_keytable[KSPACE] = newState; break;
           case SDLK_RETURN: immediate_keytable[KENTER] = newState; break;
           case SDLK_ESCAPE: immediate_keytable[KQUIT] = newState; break;
           case SDLK_KP_PLUS: case SDLK_EQUALS: immediate_keytable[KPLUS] = newState; break;
           case SDLK_KP_MINUS: case SDLK_MINUS: immediate_keytable[KMINUS] = newState; break;
           case SDLK_F1: immediate_keytable[KF1] = newState; break;
           case SDLK_F2: immediate_keytable[KF2] = newState; break;
           case SDLK_F3: immediate_keytable[KF3] = newState; break;
           case SDLK_F4: immediate_keytable[KF4] = newState; break;
           case SDLK_F5: immediate_keytable[KF5] = newState; break;
           case SDLK_F6: immediate_keytable[KF6] = newState; break;
           case SDLK_F7: immediate_keytable[KF7] = newState; break;
           case SDLK_F8: immediate_keytable[KF8] = newState; break;
           case SDLK_F9: immediate_keytable[KF9] = newState; break;
           case SDLK_F10: immediate_keytable[KF10] = newState; break;
           case SDLK_1: immediate_keytable[KNUM1] = newState; break;
           case SDLK_2: immediate_keytable[KNUM2] = newState; break;
           case SDLK_3: immediate_keytable[KNUM3] = newState; break;
           case SDLK_4: immediate_keytable[KNUM4] = newState; break;
           case SDLK_5: immediate_keytable[KNUM5] = newState; break;
           case SDLK_6: immediate_keytable[KNUM6] = newState; break;
           case SDLK_7: immediate_keytable[KNUM7] = newState; break;
           case SDLK_8: immediate_keytable[KNUM8] = newState; break;
           case SDLK_9: immediate_keytable[KNUM9] = newState; break;
           case SDLK_y: immediate_keytable[KY] = newState; break;
           case SDLK_n: immediate_keytable[KN] = newState; break;
           case SDLK_t: immediate_keytable[KT] = newState; break;
           case SDLK_TAB: immediate_keytable[KTAB] = newState; break;

           case SDLK_g: key_g = newState; break;
           case SDLK_o: key_o = newState; break;
           // key_d is set under KRIGHT2

           break;
          }

          immediate_keytable[KGOD] = (key_g & key_o & key_d);
          if (immediate_keytable[KGOD]) immediate_keytable[KRIGHT2] = 0;

          break;
        }
    }
}

